'From Squeak3.2gamma of 15 January 2002 [latest update: #4881] on 8 June 2002 at 11:02:55 am'!
"Change Set:		EdgeImage-bf
Date:			16 December 1999
Author:			Bert Freudenberg

Provides a resizable recolorable edge-preserving image morph."!

ImageMorph subclass: #EdgeImageMorph
	instanceVariableNames: 'edges drawInterior useColor useScaledEdges originalColors cachedImage '
	classVariableNames: 'DefaultEdgeForm '
	poolDictionaries: ''
	category: 'Morphic-Fur'!

!EdgeImageMorph commentStamp: '<historical>' prior: 0!
I am a resizable recolorable edge-preserving image morph. 

The image is devided into 3x3 sectors, and I warpBlt these sectors appropriately. The four corner sectors ("corners") are not scaled at all. The left and right center sectors ("edges") are scaled vertically but not horizontally, and the top and bottom edges are scaled horizontally but not vertically. The center, if visible, is scaled in both directions. So, if all edge widths are set to 0, I behave just like a scalable ImageMorph. Or, if my top edge width is set to the image height (with zero bottom edge width), I can only ever be stretched horizontally.

Recoloring me means overriding the images colors with my own background color (see Color>>#colorizeWith:). A 0.5 gray value maps to my color exactly, whereas darker/lighter grays darken/lighten my color. This is only available when the image is a ColorForm.

Inst Vars:
	edges	<Rectangle> left, right, top, and bottom edge widths
	drawInterior	<Boolean>	if false, don't draw the center sector (useful for frames)
	useColor		<Boolean>	if false, don't recolor the image
!


!Color methodsFor: 'transformations' stamp: 'bf 12/16/1999 21:29'!
colorizeWith: aColor
	"Change my color to that of aColor. Another explanation is: Darken or lighten aColor based on my brightness. Or this:
		(l _ self brightness) < 0.5 
			ifTrue:  [^ Color black mixed: (2 * l) with: aColor]
			ifFalse: [^ Color white mixed:  2 * (l-0.5) with: aColor]"
	| l r g b |
	l _ (self privateRed + self privateGreen + self privateBlue) // 3.
	l <= HalfComponentMask
		ifTrue: [r _ (aColor privateRed * l) // HalfComponentMask.
				g _ (aColor privateGreen * l) // HalfComponentMask.
				b _ (aColor privateBlue * l) // HalfComponentMask]
		ifFalse: [l _ ComponentMask - 1 - ((l - HalfComponentMask) * 2).
				r _ ComponentMask - (((ComponentMask - aColor privateRed) * l) // ComponentMask).
				g _ ComponentMask - (((ComponentMask - aColor privateGreen) * l) // ComponentMask).
				b _ ComponentMask - (((ComponentMask - aColor privateBlue) * l) // ComponentMask)].
	^Color basicNew setPrivateRed: r green: g blue: b! !


!FormCanvas methodsFor: 'other' stamp: 'jcg 5/30/2002 00:40'!
warp: aForm
	"Answer a WarpBlt that can be used to warp stuff from aForm to me. Clients still need to offset destRect by my origin!!"
	| warp |
	warp _ (WarpBlt toForm: form)
 		sourceForm: aForm;
		clipRect: clipRect;
		combinationRule: Form paint.
	self isShadowDrawing
		ifTrue: [warp colorMap: (Color maskingMap: aForm depth);
					fillColor: shadowColor]
		ifFalse: [warp colorMap: (aForm colormapIfNeededForDepth: form depth);
					fillColor: nil].
	^warp! !


!ImageMorph methodsFor: 'initialization' stamp: 'bf 12/16/1999 20:38'!
initialize

	super initialize.
	self image: self class defaultForm.
! !


!EdgeImageMorph methodsFor: 'initialization' stamp: 'jcg 6/2/2002 12:54'!
initialize
	super initialize.
	edges _ Rectangle left: 4 right: 4 top: 4 bottom: 4.
	drawInterior _ true.
	useColor _ false.
	useScaledEdges _ true

! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 12/16/1999 20:14'!
color: aColor
	color = aColor ifFalse: [
		color _ aColor.
		self recolorImage.
		self changed]! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 12/16/1999 20:56'!
drawInterior: aBoolean
	drawInterior = aBoolean ifFalse:
		[drawInterior _ aBoolean.
		self changed]! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 12/14/1999 12:56'!
edgeLeft: left right: right top: top bottom: bottom
	self setEdges: (Rectangle left: left right: right top: top bottom: bottom)! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 12/14/1999 12:53'!
edgeWidth: width
	self edgeLeft: width right: width top: width bottom: width! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 1/12/2001 16:54'!
fillStyle: aFillStyle
	| oldColor |
	oldColor _ color.
	super fillStyle: aFillStyle.
	color = oldColor ifFalse: [
		self recolorImage.
		self changed]! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2002 11:01'!
image: anImage
	| oldImageExtent |
	oldImageExtent _ image ifNotNil: [image extent].
	anImage depth = 1
		ifTrue: [image _ ColorForm mappingWhiteToTransparentFrom: anImage]
		ifFalse: [image _ anImage].
	originalColors _ (image isKindOf: ColorForm) ifTrue: [
		image _ image shallowCopy.
		image colors].
	self recolorImage.
	oldImageExtent = image extent
		ifFalse: [self extent: image extent].
	cachedImage _ nil.
	self changed.

! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'bf 12/16/1999 21:00'!
useColor: aBoolean
	useColor = aBoolean ifFalse: [
		useColor _ aBoolean.
		originalColors notNil ifTrue:
			[useColor
				ifTrue: [self recolorImage]
				ifFalse: [image colors: originalColors].
			self changed]]! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'jcg 6/2/2002 12:55'!
useScaledEdges

	^ useScaledEdges! !

!EdgeImageMorph methodsFor: 'accessing' stamp: 'jcg 6/2/2002 12:55'!
useScaledEdges: aBoolean
	useScaledEdges = aBoolean ifFalse: [
		useScaledEdges _ aBoolean.
		cachedImage _ nil.
		self changed]! !

!EdgeImageMorph methodsFor: 'drawing' stamp: 'jcg 6/2/2002 12:51'!
bltScaledEdges: warp outerSrc: outSrc innerSrc: inSrc outerDest: outDest innerDest: inDest
	| centerValid |

	centerValid _ true.

	"Top/bottom edges"
	(inSrc width > 0 and: [inDest width > 0]) ifTrue:
		[edges top > 0 ifTrue: 
			[warp copyQuad: (inSrc left@outSrc top corner: inSrc topRight) innerCorners
				toRect: (inDest left@outDest top corner: inDest topRight)].
		edges bottom > 0 ifTrue:
			[warp copyQuad: (inSrc bottomLeft corner: inSrc right@outSrc bottom) innerCorners
				toRect: (inDest bottomLeft corner: inDest right@outDest bottom)]]
		ifFalse: [centerValid _ false].
	"Left/right edges"
	(inSrc height > 0 and: [inDest height > 0]) ifTrue:
		[edges left > 0 ifTrue:
			[warp copyQuad: (outSrc left@inSrc top corner: inSrc bottomLeft) innerCorners
				toRect: (outDest left@inDest top corner: inDest bottomLeft)].
		edges right > 0 ifTrue:
			[warp copyQuad: (inSrc topRight corner: outSrc right@inSrc bottom) innerCorners
				toRect: (inDest topRight corner: outDest right@inDest bottom)]]
		ifFalse: [centerValid _ false].

	drawInterior & centerValid ifTrue: [
		warp copyQuad: inSrc innerCorners toRect: inDest]! !

!EdgeImageMorph methodsFor: 'drawing' stamp: 'jcg 6/2/2002 12:50'!
bltTiledEdges: canvas outerSrc: outSrc innerSrc: inSrc outerDest: outDest innerDest: inDest
	| block |

	block _ [:srcRect :destRect |
		(srcRect hasPositiveExtent and: [destRect hasPositiveExtent]) ifTrue: [
			canvas 
				infiniteFillRectangle: destRect
				fillStyle: (InfiniteForm with: (image contentsOfArea: srcRect))]].

	block value: self topEdgeSrcRect value: self topEdgeDestRect.
	block value: self bottomEdgeSrcRect value: self bottomEdgeDestRect.
	block value: self leftEdgeSrcRect value: self leftEdgeDestRect.
	block value: self rightEdgeSrcRect value: self rightEdgeDestRect.
	drawInterior ifTrue: [
		block value: self centerSrcRect value: self centerDestRect].! !

!EdgeImageMorph methodsFor: 'drawing' stamp: 'jlb 6/8/2002 11:02'!
drawOn: aCanvas
	cachedImage ifNil: [self computeCachedImage].
	self isOpaque
		ifTrue:[aCanvas drawImage: cachedImage at: self position]
		ifFalse:[aCanvas paintImage: cachedImage at: self position]! !

!EdgeImageMorph methodsFor: 'geometry' stamp: 'jcg 6/2/2002 11:00'!
extent: aPoint
	| min ext |
	(image isNil or: [edges isNil]) ifTrue: [^super extent: aPoint].
	"Constrain minimal extent to sum of edge widths"
	min _ self minWidth @ self minHeight.
	ext _ aPoint max: min.
	"Constrain maximal extent if no stretchable interior"
	image width > min x ifFalse: [ext _ min x @ ext y].
	image height > min y ifFalse: [ext _ ext x @ min y].
	bounds extent = ext ifFalse: [
		self changed.
		bounds _ bounds topLeft extent: ext.
		cachedImage _ nil.
		self layoutChanged.
		self changed].
! !

!EdgeImageMorph methodsFor: 'geometry' stamp: 'bf 12/15/1999 06:04'!
innerBounds
	^bounds insetBy: edges! !

!EdgeImageMorph methodsFor: 'geometry' stamp: 'bf 12/15/1999 01:11'!
minHeight
	^edges top + edges bottom
! !

!EdgeImageMorph methodsFor: 'geometry' stamp: 'bf 12/15/1999 01:08'!
minWidth
	^edges left + edges right
! !

!EdgeImageMorph methodsFor: 'menu commands' stamp: 'jcg 6/2/2002 12:57'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	useColor
		ifTrue: [aCustomMenu add: 'original image colors' action: #toggleUseColor]
		ifFalse: [aCustomMenu add: 'colorize image' action: #toggleUseColor].
	drawInterior
		ifTrue: [aCustomMenu add: 'draw edges only' action: #toggleDrawInterior]
		ifFalse: [aCustomMenu add: 'draw interior' action: #toggleDrawInterior].
	useScaledEdges
		ifTrue: [aCustomMenu add: 'draw tiled edges' action: #toggleScaledEdges]
		ifFalse: [aCustomMenu add: 'draw scaled edges' action: #toggleScaledEdges].
! !

!EdgeImageMorph methodsFor: 'menu commands' stamp: 'bf 12/16/1999 21:41'!
toggleDrawInterior
	self drawInterior: drawInterior not.
! !

!EdgeImageMorph methodsFor: 'menu commands' stamp: 'jcg 6/2/2002 12:54'!
toggleScaledEdges
	self useScaledEdges: useScaledEdges not.
! !

!EdgeImageMorph methodsFor: 'menu commands' stamp: 'bf 12/16/1999 21:41'!
toggleUseColor
	self useColor: useColor not.
! !

!EdgeImageMorph methodsFor: 'private' stamp: 'jcg 6/2/2002 12:52'!
computeCachedImage
	| canvas outerSrc innerSrc outerDest innerDest warp |

	cachedImage _ Form extent: self extent depth: image depth.
	canvas _ FormCanvas on: cachedImage.

	outerSrc _ image boundingBox.
	innerSrc _ outerSrc insetBy: edges.
	outerDest _ 0@0 corner: self extent.
	innerDest _ outerDest insetBy: edges.

	warp _ canvas warp: image.
	"Top corners"
	edges top > 0 ifTrue:
		[edges left > 0 ifTrue: 
			[warp copyQuad: (outerSrc topLeft corner: innerSrc topLeft) innerCorners
				toRect: (outerDest topLeft corner: innerDest topLeft)].
		edges right > 0 ifTrue: 
			[warp copyQuad: (outerSrc topRight rect: innerSrc topRight) innerCorners
				toRect: (outerDest topRight rect: innerDest topRight)]].
	"Bottom corners"
	edges bottom > 0 ifTrue:
		[edges left > 0 ifTrue: 
			[warp copyQuad: (outerSrc bottomLeft rect: innerSrc bottomLeft) innerCorners
				toRect: (outerDest bottomLeft rect: innerDest bottomLeft)].
		edges right > 0 ifTrue: 
			[warp copyQuad: (innerSrc bottomRight corner: outerSrc bottomRight) innerCorners
				toRect: (innerDest bottomRight corner: outerDest bottomRight)]].

	self useScaledEdges 
		ifTrue: [self 
					bltScaledEdges: warp 
					outerSrc: outerSrc 
					innerSrc: innerSrc 
					outerDest: outerDest
					innerDest: innerDest]
		ifFalse: [self 
					bltTiledEdges: canvas 
					outerSrc: outerSrc 
					innerSrc: innerSrc 
					outerDest: outerDest
					innerDest: innerDest].

	! !

!EdgeImageMorph methodsFor: 'private' stamp: 'bf 12/16/1999 20:58'!
recolorImage
	"Recolor image based on it's saved colors"
	(useColor == true and: [originalColors notNil
		and: [color isColor and: [color isTransparent not]]])
			ifTrue: [image colors: (originalColors collect:
						[:col | col colorizeWith: color])]! !

!EdgeImageMorph methodsFor: 'private' stamp: 'bf 12/14/1999 14:52'!
setEdges: aRectangle

	edges _ aRectangle.
	self changed.
	self layoutChanged.
	self extent: self extent! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:36'!
bottomEdgeDestRect

	^ edges left@(self height - edges bottom) corner: (self width - edges right)@self height! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:36'!
bottomEdgeSrcRect

	^ edges left@(image height - edges bottom) corner: (image width - edges right)@image height! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:48'!
centerDestRect

	^ edges left@edges top corner: (self width - edges right)@(self height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:48'!
centerSrcRect

	^ edges left@edges top corner: (image width - edges right)@(image height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:39'!
leftEdgeDestRect

	^ 0@edges top corner: edges left@(self height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:41'!
leftEdgeSrcRect

	^ 0@edges top corner: edges left@(image height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:42'!
rightEdgeDestRect

	^ (self width - edges right)@edges top corner: self width@(self height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:42'!
rightEdgeSrcRect

	^ (image width - edges right)@edges top corner: image width@(image height - edges bottom)! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:34'!
topEdgeDestRect

	^ edges left@0 corner: (self width - edges right)@edges top! !

!EdgeImageMorph methodsFor: 'private-accessing' stamp: 'jcg 6/2/2002 12:34'!
topEdgeSrcRect

	^ edges left@0 corner: (image width - edges right)@edges top! !


!EdgeImageMorph class methodsFor: 'class initialization' stamp: 'bf 12/16/1999 20:51'!
initialize
	"EdgeImageMorph initialize"

	DefaultEdgeForm _ (ColorForm
		extent: 64@32
		depth: 8
		fromArray: #( 4294967295 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124286 4278124287 4294967295 4294965233 4025413358 3991793133 3974950123 3958106858 3941263849 3924355304 3907512295 3890669286 3873826277 3856983268 3840205795 3823363043 3823297250 3806520034 2842755071 4294438888 3772767966 3722239195 3688552921 3638089686 3604338132 3553874642 3520188624 3486502606 3452816588 3419130570 3402222024 3368601799 3351758790 3334915782 2489528575 4294109664 3503345357 3452750538 3385378758 3318006978 3250700479 3200105659 3132799416 3082270389 3031741362 2981212335 2930683565 2896997291 2880154282 2846468521 2119379455 4277133521 3250634686 3166419640 3082204596 2997989551 2913774506 2846402213 2762187425 2694815389 2627443353 2560071317 2492699538 2442236047 2408484493 2374864012 1766204160 4277133520 3250634429 3166353848 3082138803 2997923758 2913708713 2829559460 2745344416 2677972380 2610600344 2543228308 2475856273 2425327502 2374864012 2357955467 1749361152 4277133520 3233791676 3149576631 3065361330 2981146285 2896931496 2812716452 2728501407 2661129371 2593757335 2526385299 2459013264 2408484237 2358021003 2324335241 1732518400 4277133264 3233725884 3132733623 3048518578 2964303532 2880088488 2795873443 2728435614 2644286362 2576914326 2509542290 2442170254 2391641228 2341112457 2307491976 1715741184 4277133264 3216948667 3132733366 3048452785 2964237740 2880022695 2795807650 2711658141 2627443353 2560071317 2492699281 2425327245 2358021002 2324269192 2290583431 1682120960 4277133263 3216882875 3115890613 3031675568 2947460267 2863245222 2779030177 2694815389 2610600344 2543228308 2475856272 2408484236 2341112457 2290648967 2256962949 1682055424 4277067727 3200105658 3115890357 3031609776 2930617515 2846402469 2762187425 2677972380 2610534551 2526385299 2459013263 2391641227 2324269192 2273740421 2240054404 1665212416 4277067470 3200039866 3099047604 3014832303 2930617258 2846336677 2762121632 2677906587 2593757078 2509542290 2442170254 2374798218 2307426183 2256897412 2223211394 1631657984 4277067470 3183262649 3099047348 3014766767 2913774505 2829559460 2745344159 2661129114 2576914325 2492699281 2425327245 2357955209 2290583173 2240054147 2189590913 1631592448 4277067469 3183262393 3082204595 2997989294 2913708713 2829493668 2728501407 2644286362 2560071317 2475856272 2408484236 2341112200 2273740164 2206433921 2172682368 1597972224 4277067469 3166419640 3082204339 2997923758 2896931496 2812716195 2728435614 2644220569 2560005524 2475790479 2391641227 2324269191 2256897155 2189525376 2155839358 1581195008 4277067469 3166419384 3082138803 2981146541 2896865704 2795873443 2711658397 2627443096 2543228051 2459013262 2374798218 2307426182 2240054146 2172682367 2122218877 1581129472 4277067213 3166353848 3065361586 2981080749 2880088487 2795873186 2711592605 2610600344 2526385298 2442170254 2357955209 2290583173 2223211137 2155839102 2105310331 1547509248 4277067213 3149576631 3065361330 2964303532 2880088231 2795807650 2694815388 2610600087 2526319506 2442104461 2357889416 2273740164 2206368128 2138996092 2088467322 1530732032 4277067212 3149576375 3065295793 2964303276 2880022694 2779030177 2694749596 2593757334 2509542289 2425326988 2341111943 2256897155 2189525119 2122153083 2054846841 1513889280 4277067212 3149510839 3048518321 2964237740 2863245222 2778964641 2677972379 2593757078 2509476497 2408484236 2324269191 2240054146 2172682110 2105310074 2038003831 1497046272 4277067212 3149510838 3048518321 2947460523 2863179686 2762187424 2677972123 2576914325 2492699024 2408418443 2324203398 2239988353 2155839101 2088467065 2021095286 1480203520 4277001676 3132733366 3048452785 2947460267 2863179685 2762187168 2677906586 2576914069 2492633488 2391641226 2307425925 2223210880 2138996092 2071624056 2004252021 1463426304 4277001675 3132733366 3031675568 2947394731 2846402213 2762121632 2661129114 2576848533 2475856271 2391575434 2290583173 2206368128 2122153083 2054781047 1987409011 1446583296 4277001675 3132667830 3031675312 2947394730 2846402213 2745344415 2661129114 2560071316 2475790479 2374798217 2290582916 2206302335 2122087290 2037938038 1970566002 1429740544 4277001419 3132667829 3031675312 2930617514 2846336677 2745344159 2661063577 2560071060 2475790478 2374797961 2290517380 2189525118 2105309817 2021095029 1953722993 1412963328 4277001419 3115890613 3031609776 2930617258 2846336676 2745344159 2644286361 2560005524 2459013006 2374732425 2273739907 2189459326 2088467065 2004252020 1936879984 1396120320 4277001419 3115890613 3031609775 2930617258 2829559460 2745278623 2644286105 2560005523 2459013006 2357955208 2273674371 2172682109 2088466808 2004186227 1920036975 1379277568 4277001419 3115890357 3031609775 2930617258 2829559204 2745278622 2644286105 2543228307 2458947470 2357954952 2273674370 2172681853 2071624055 1987408754 1903193966 1379277312 4289897880 2341112200 2273740163 2206302336 2122153083 2054781046 1987343219 1903193966 1835821930 1768449894 1684300641 1633640030 1549490777 1482118741 1414746706 1026036991 4285951589 1549556570 1499027543 1465275733 1414746706 1364217935 1313688908 1263225417 1212696390 1178944580 1128415553 1077886782 1027423291 976894264 943142454 672861695 4294915634 774778157 741092395 724249130 707340585 673720359 656811558 623191332 606348067 589439522 555819040 538910495 505290269 488447004 471538458 336461823 4294967295 0 0 0 0 0 0 0 0 0 0 0 0 0 0 4294967295)
	offset: 0@0)
	colors: ((1 to: 256) collect: [:i | i < 256
		ifTrue: [Color gray: i / 255.0]
		ifFalse: [Color transparent]]).

! !

!EdgeImageMorph class methodsFor: 'accessing' stamp: 'bf 12/16/1999 20:52'!
defaultForm
	^DefaultEdgeForm! !


!TranslucentColor methodsFor: 'accessing' stamp: 'bf 12/15/1999 08:22'!
colorizeWith: aColor
	^(super colorizeWith: aColor) alpha: self alpha! !

EdgeImageMorph initialize!
EdgeImageMorph removeSelector: #drawEdgeImageOn:!

!EdgeImageMorph reorganize!
('initialization' initialize)
('accessing' color: drawInterior: edgeLeft:right:top:bottom: edgeWidth: fillStyle: image: useColor: useScaledEdges useScaledEdges:)
('drawing' bltScaledEdges:outerSrc:innerSrc:outerDest:innerDest: bltTiledEdges:outerSrc:innerSrc:outerDest:innerDest: drawOn:)
('geometry' extent: innerBounds minHeight minWidth)
('menu commands' addCustomMenuItems:hand: toggleDrawInterior toggleScaledEdges toggleUseColor)
('private' computeCachedImage recolorImage setEdges:)
('private-accessing' bottomEdgeDestRect bottomEdgeSrcRect centerDestRect centerSrcRect leftEdgeDestRect leftEdgeSrcRect rightEdgeDestRect rightEdgeSrcRect topEdgeDestRect topEdgeSrcRect)
!

