'From Squeak3.2gamma of 15 January 2002 [latest update: #4743] on 1 March 2002 at 7:52:23 pm'!
EdgeImageMorph subclass: #ModeButtonMorph
	instanceVariableNames: 'state target actionSelector arguments actWhen '
	classVariableNames: 'AuthorModeOwner '
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!ModeButtonMorph commentStamp: '<historical>' prior: 0!
I represent an abstract class for fixed size bitmapped image buttons.

My subclasses represent buton morphs with seperate images for on, off and pressed with the mouse.

When the event actWhen occurs, send actionSelector with 'arguments' to target.  For other events, default to my eventHandler.  The current event is not supplied in the arguments to the actionSelector.  

There is a lot of error checking on cursor setting because 'nil beCursor' tends to have some negative side effects !

ModeButtonMorph subclass: #TwoModeButtonMorph
	instanceVariableNames: 'active disabled forms '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!TwoModeButtonMorph commentStamp: '<historical>' prior: 0!
I represent a button morph with seperate images for different button states.

When the event actWhen occurs, I send actionSelector with 'arguments' to target.  For other events, I default to my eventHandler.  The current event is not supplied in the arguments to the actionSelector.  

I have two instance variables

active --  Boolean  - the button is active (true) or passive ( false ) 
		E.g. This is useful when used in a window title when the title bar goes from active to passive

disabled -- the button is disabled

Explanation:

A TwoModeButtonMorph provides a 'regular' OS type of window button. Typically these are used as the close button, the expand button and the collapse button. Different types of window may have more or less buttons.

A window is either in one of two states. When the window is topmost, or has focus, we call that the active state. When the window is in the background, or loses focus, we call that the passive state.

Belying its name, the TwoModeButtonMorph works as follows. When the window that the button belongs is active, the button displays an active image. When pressed, the button displays a pressed-active image. In addition, the button may be disabled, in which case the button will always display the disabled-active image, even when pressed. 

When the window that the button belongs to is passive, the button displays a passive image. When pressed, the button displays a pressed-passive image. The button may be in a disabled state in passive mode also, under which circumstances it displays the disabled-passive.

Though named a TwoModeButtonMorph, there are really six different states that the morph can be in. The "two mode" refers to the fact that within its interaction mode, it has a "normal" and "pressed" mode.

Typically each of these images is represented as a bitmap, These images are stored as six properties:

activeImage
passiveImage
pressedActiveImage
pressedPassiveImage
disabledActiveImage
disabledPassiveImage


Instead of providing images, a renderCode block may be provided. The block is called from the drawOn: method, and has two parameters. The first parameter is the TwoModeButtonMorph itself; the second is the canvas on which to draw the morph. Here's an example:

bo :=  TwoModeButtonMorph new.
bo renderCode: [ :morph :canvas |  canvas fillRectangle: morph bounds fillStyle: morph  fillStyle ].
bo openInWorld.

Of course, this only draws one style; a proper button would draw all six different button states correctly.

Also, the button has an inherited instance variable cursor, which defines what the shape of the cursor (also known as the hand, an instance of HandMorph) should be when within the bounds of the button. In the case that the hand is carrying something, cursor switching is ignored. The default cursor for a TwoModeButtonMorph is Cursor normal.

A proper button press sequence (mouse down, mouse up within the bounds of the button) triggers a user-defined message.

!
]style[(523 12 2427)f1,f2,f1!

TwoModeButtonMorph subclass: #OSSystemButtonMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zurgle-Controls'!

!OSSystemButtonMorph commentStamp: '<historical>' prior: 0!
I represent an imaged button that reponds to the mouse with four different forms:

	1. Normal
	2. Mouse down
	3. Mouse over
	4. Disabled!

TwoModeButtonMorph class
	instanceVariableNames: ''!

!ModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 6/8/2001 13:09'!
adaptToWorld: aWorld
	super adaptToWorld: aWorld.
	self target: (target adaptedToWorld: aWorld).! !

!ModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 7/10/2001 23:15'!
defaultRenderBlock
^  [ :morph :canvas | canvas fillRectangle: morph bounds fillStyle: morph fillStyle ]! !

!ModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 7/10/2001 23:17'!
defaultRenderCode
^  [ :morph :canvas | canvas fillRectangle: morph bounds fillStyle: morph fillStyle ]! !

!ModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 9/8/2001 15:20'!
initialize

	super initialize.
	state _ #off.
	target _ nil.
	actionSelector _ #flash.
	arguments _ EmptyArray.
	actWhen _ #buttonUp.
	self color: Color yellow.
! !

!ModeButtonMorph methodsFor: 'copying' stamp: 'jlb 6/8/2001 13:11'!
updateReferencesUsing: aDictionary
	"If the arguments array points at a morph we are copying, then update it to point to the new copy. This method also copies the arguments array itself, which is important!!"

	super updateReferencesUsing: aDictionary.
	arguments _ arguments collect:
		[:old | aDictionary at: old ifAbsent: [old]].
! !

!ModeButtonMorph methodsFor: 'copying' stamp: 'jlb 6/8/2001 13:11'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target _ deepCopier references at: target ifAbsent: [target].
arguments _ arguments collect: [:each |
	deepCopier references at: each ifAbsent: [each]].
! !

!ModeButtonMorph methodsFor: 'copying' stamp: 'jlb 7/22/2001 19:56'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
state _ state veryDeepCopyWith: deepCopier.
"target _ target.		Weakly copied"
"actionSelector _ actionSelector.		Symbol"
"arguments _ arguments.		Weakly copied"
actWhen _ actWhen.		"Symbol"
! !

!ModeButtonMorph methodsFor: 'geometry' stamp: 'jlb 7/11/2001 22:32'!
morphExtent: aPoint
	" this should be the same as Morph>>extent: "

	bounds extent = aPoint ifTrue: [^ self].
	self changed.
	bounds _ bounds topLeft extent: aPoint.
	self layoutChanged.
	self changed.
! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 7/11/2001 14:46'!
cursorReset

	| aCursor |
	aCursor _ self valueOfProperty: #savedCursor ifAbsent: [ Cursor normal ].
	aCursor ifNil: [ Cursor normal show ]
		ifNotNil: [ aCursor show].
	self removeProperty: #savedCursor.
! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 9/8/2001 11:41'!
doButtonAction
	"Perform the action of this button. Subclasses may override this method. The default behavior is to send the button's actionSelector to its target object with its arguments."

	(target ~~ nil and: [actionSelector ~~ nil]) ifTrue: [
		Cursor normal showWhile: [
			target perform: actionSelector withArguments: arguments].
		target isMorph ifTrue: [target changed]].
! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:12'!
doButtonAction: evt
	| moreArgs |
	"Perform the action of this button. Subclasses may override this method. The default behavior is to send the button's actionSelector to its target object with its arguments."

	target ifNil: [^self].
	actionSelector ifNil: [^self].
	Cursor normal showWhile: [
		moreArgs _ actionSelector numArgs > arguments size ifTrue: [
			arguments copyWith: evt
		] ifFalse: [
			arguments
		].
		target perform: actionSelector withArguments: moreArgs
	]! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:12'!
handlesMouseDown: evt

	^ true
! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:12'!
handlesMouseStillDown: evt
	^actWhen == #whilePressed! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:12'!
mouseDown: evt
	| now dt |
	self state: #pressed.
	actWhen == #buttonDown
		ifTrue:
			[self doButtonAction]
		ifFalse:
			[now _ Time millisecondClockValue.
			super mouseDown: evt.
			"Allow on:send:to: to set the response to events other than actWhen"
			dt _ Time millisecondClockValue - now max: 0.  "Time it took to do"
			dt < 200 ifTrue: [(Delay forMilliseconds: 200-dt) wait]].
	self mouseStillDown: evt.! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 7/11/2001 14:30'!
mouseLeave: event

	self cursorReset! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:12'!
mouseStillDown: evt
	actWhen == #whilePressed ifFalse:[^self].
	(self containsPoint: evt cursorPoint) ifTrue:[self doButtonAction].! !

!ModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 7/11/2001 14:41'!
mouseUp: evt
	"Allow on:send:to: to set the response to events other than actWhen"
	actWhen == #buttonUp ifFalse: [^super mouseUp: evt].

	(self containsPoint: evt cursorPoint) ifTrue: [
		self state: #on.
		self doButtonAction: evt
	] ifFalse: [
		self state: #off.
		target ifNotNil: [target mouseUpBalk: evt].
		self cursorReset.
	].
! !

!ModeButtonMorph methodsFor: 'stepping' stamp: 'jlb 9/8/2001 15:07'!
step
	(self hasProperty: #doesButtonAction) ifTrue:[
		self doButtonAction.
		self setProperty: #didButtonAction toValue: true.
	].! !

!ModeButtonMorph methodsFor: 'stepping' stamp: 'jlb 6/8/2001 13:13'!
stepTime
	(self hasProperty: #doesButtonAction) ifTrue:[^1].
	^super stepTime! !

!ModeButtonMorph methodsFor: 'stepping' stamp: 'jlb 4/20/2002 21:26'!
wantsSteps
	^ false.
"
	^(self hasProperty: #doesButtonAction) or:[super wantsSteps]" ! !

!ModeButtonMorph methodsFor: 'printing' stamp: 'jlb 6/8/2001 13:13'!
printOn: aStream

	| string |

	aStream nextPutAll: 'ModeButton'.
	arguments size > 0 ifTrue: [string _ arguments at: (2 min: arguments size)].
	aStream nextPutAll: '('.
	(string ~~ nil and: [string ~~ self])
		ifTrue:
			[aStream print: string; space]
		ifFalse:
			[aStream print: actionSelector; space].
	aStream print: self identityHash;
			nextPutAll: ')'.! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:09'!
actWhen: condition
	"Accepts symbols:  #buttonDown, #buttonUp, and #whilePressed"
	actWhen _ condition! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:09'!
actionSelector

	^ actionSelector
! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:09'!
actionSelector: aSymbolOrString

	(nil = aSymbolOrString or:
	 ['nil' = aSymbolOrString or:
	 [aSymbolOrString isEmpty]])
		ifTrue: [^ actionSelector _ nil].

	actionSelector _ aSymbolOrString asSymbol.
! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:09'!
arguments
	^ arguments! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
arguments: aCollection

	arguments _ aCollection asArray copy.
! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
authorModeOwner: aMorph
	AuthorModeOwner _ aMorph! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
dragIfAuthoring: evt
	"Allow simple dragging if the class var is set to my owner."
	owner == AuthorModeOwner ifTrue: [
		self center: evt cursorPoint].
	^ owner == AuthorModeOwner! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
isOn
	^ state == #on! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 7/11/2001 22:54'!
state
	^ state! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
state: newState
	"Change the image and invalidate the rect."

	newState == state ifTrue: [^ self].
	state _ newState.
	self invalidRect: bounds.	"All three images must be the same size"! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
target

	^ target
! !

!ModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:10'!
target: anObject

	target _ anObject
! !


!ModeButtonMorph class methodsFor: 'form handling' stamp: 'jlb 9/15/2001 20:19'!
readFormAndFix: aFileNameString replaceColor: aColor 
	"read a Form from a file and replace aColor with Color transparent"
	| aForm |
	aForm _ Form fromFileNamed: aFileNameString.
	aForm depth < 16
		ifTrue: [aForm _ aForm asFormOfDepth: 32].
	aForm replaceColor: aColor withColor: Color transparent.
	^ aForm! !


!TwoModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 10/7/2001 09:59'!
from4Images: buttonImages 
	" bogus ; makes 4 images appear as 6 in TwoModeButtonMorph "
	
	| bogus |
	self fromImages: buttonImages symbols: #( #off #pressed #offPassive #disabled ).
	" now finish filling out "
	bogus _ forms at: #disabled.
	forms at: #pressedPassive put: bogus.
	forms at: #disabledPassive put: bogus.
	self state: #off.
! !

!TwoModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 9/24/2001 12:02'!
fromImages: buttonImages 
	"assume that all six buttons are there vertically stacked"

	self fromImages: buttonImages symbols: #( #off #pressed #disabled #offPassive #pressedPassive #disabledPassive ).
	self state: #off.
! !

!TwoModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 9/24/2001 11:43'!
fromImages: buttonImages symbols: anArrayOfSymbols
	"assume that there are (anArrayOfSymbols size) buttons vertically stacked"
	" button images contains the form image"

	| formExtent height aForm buttonExtent noImages |

	noImages _ anArrayOfSymbols size.
	formExtent _ buttonImages extent.
	height _ formExtent y / noImages.
	buttonExtent _ formExtent x @ height.

	0 to: (noImages-1)
		do: [:i | 
			aForm _ buttonImages
						contentsOfArea: (0 @ (i * height) extent: buttonExtent).
			forms at: ( anArrayOfSymbols at: ( i+1)) put: aForm.
		 ].
! !

!TwoModeButtonMorph methodsFor: 'initialization' stamp: 'jlb 9/24/2001 12:01'!
initialize
	super initialize.
	active _ true.
	disabled _ false.
	forms _ Dictionary new.
	self setProperty: #wantsActivate toValue: true .

	! !

!TwoModeButtonMorph methodsFor: 'accessing'!
activate
	self active: true.
! !

!TwoModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:32'!
active
	^ active! !

!TwoModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 7/11/2001 23:05'!
active: aBoolean
	active _ aBoolean.
	self state: state.
	self changed! !

!TwoModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 6/8/2001 13:43'!
disabled
	^ disabled! !

!TwoModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 7/11/2001 23:05'!
disabled: aBoolean
	disabled _ aBoolean.
	self state: state.
	self changed.! !

!TwoModeButtonMorph methodsFor: 'accessing'!
passivate
	self active: false.
! !

!TwoModeButtonMorph methodsFor: 'accessing' stamp: 'jlb 9/24/2001 11:56'!
state: newState 
	"Change the image and invalidate the rect."
	| symbol |

	state _ newState.
	symbol _ #off.
	symbol _ disabled
				ifTrue: [#disabled]
				ifFalse: [state].
	active
		ifFalse: [symbol _ (symbol , 'Passive') asSymbol].
	self
		image: (forms at: symbol)! !

!TwoModeButtonMorph methodsFor: 'event handling'!
mouseDown: evt 
	disabled
		ifTrue: [^ self].
	"get rid of the inherited hesitation; don't call super mouseDown: evt"
	self state: #pressed.
	active
		ifFalse: [^ self].
	actWhen == #buttonDown
		ifTrue: [self doButtonAction]
		ifFalse: [
			" Morph>>mouseDown: ; this is useful if you need to get a hold of the event "
			self eventHandler
				ifNotNil: [self eventHandler mouseDown: evt fromMorph: self]].
	self mouseStillDown: evt! !

!TwoModeButtonMorph methodsFor: 'event handling' stamp: 'je 7/31/2001 19:24'!
mouseMove: evt

	(active and: [ disabled not ]) 
		ifTrue: [ self adjustState: evt. super mouseMove: evt ].! !

!TwoModeButtonMorph methodsFor: 'event handling' stamp: 'jlb 6/8/2001 13:50'!
mouseStillDown: evt

	(active and: [ disabled not ]) 
		ifTrue: [ super mouseStillDown: evt ].! !

!TwoModeButtonMorph methodsFor: 'event handling'!
mouseUp: evt 
	disabled
		ifFalse: ["reset the images"
			self state: #off.
			active
				ifFalse: [^ self].
			(self containsPoint: evt cursorPoint)
				ifTrue: ["Allow on:send:to: to set the response to events 
					other  
					than actWhen"
					actWhen == #buttonUp
						ifTrue: [ self doButtonAction: evt]
						ifFalse: [ " Morph>>mouseDown: ; this is useful if you need to get a hold of the event "
								  super mouseUp: evt]]
				ifFalse: [target
						ifNotNil: [target mouseUpBalk: evt].
					self cursorReset]]! !

!TwoModeButtonMorph methodsFor: 'visual properties' stamp: 'je 7/31/2001 19:23'!
adjustState: evt
	"while pressed, only show the pressed image while the hand is still over the button"

	| desiredState |
     desiredState _ (self containsPoint: evt cursorPoint)
		ifTrue: [ #pressed ]
		ifFalse: [ #off ].

     (self state ~= desiredState) ifTrue: [ self state: desiredState. self changed ].! !

!TwoModeButtonMorph methodsFor: 'copying' stamp: 'jlb 9/24/2001 12:13'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
active _ active veryDeepCopyWith: deepCopier.
disabled _ disabled copy.
forms _ forms veryDeepCopyWith: deepCopier.
! !

!TwoModeButtonMorph methodsFor: 'testing'!
isPressed
	^ state == #pressed! !


!OSSystemButtonMorph methodsFor: 'event handling' stamp: 'jlb 2/27/2002 21:08'!
mouseEnter: evt for: theMorph
	self state: #focus! !

!OSSystemButtonMorph methodsFor: 'event handling' stamp: 'jlb 2/27/2002 21:24'!
mouseLeave: evt 
	super mouseLeave: evt.
	self state: #off
! !

!OSSystemButtonMorph methodsFor: 'event handling' stamp: 'jlb 2/27/2002 22:04'!
mouseUp: evt 
	disabled
		ifFalse: ["reset the images"
			self state: #off.
			active
				ifFalse: [^ self].
			(self containsPoint: evt cursorPoint)
				ifTrue: ["Allow on:send:to: to set the response to events  
					other  
					than actWhen"
					self state: #focus.
					actWhen == #buttonUp
						ifTrue: [self doButtonAction: evt]
						ifFalse: ["Morph>>mouseDown: ; this is useful if you 
							need to get a hold of the event"
							super mouseUp: evt]]
				ifFalse: [target
						ifNotNil: [target mouseUpBalk: evt].
					self cursorReset]]! !

!OSSystemButtonMorph methodsFor: 'event handling' stamp: 'jlb 2/27/2002 23:06'!
state: aStateSymbol 
	"Change the image and invalidate the rect."
	| symbol |
	state _ aStateSymbol.
	symbol _ #normal.
	symbol _ active
				ifFalse: [disabled
						ifTrue: [#disabledInactive]
						ifFalse: [state caseOf: {
							[#off] -> [#normalInactive].
							[#pressed] -> [#pressedInactive].
							[#focus] -> [#hotInactive]}]]
				ifTrue: [disabled
						ifTrue: [#disabled]
						ifFalse: [state caseOf: {
							[#off] -> [#normal]. 
							[#pressed] -> [#pressed]. 
							[#focus] -> [#hot]}]].
	self
		image: (forms at: symbol)! !

!OSSystemButtonMorph methodsFor: 'initialization' stamp: 'jlb 2/27/2002 21:23'!
fromImages: buttonImagesForm 
	self fromImages: buttonImagesForm symbols: #(#normal #hot #pressed #disabled #normalInactive #hotInactive #pressedInactive #disabledInactive ).
	self state: #off.! !

!OSSystemButtonMorph methodsFor: 'initialization' stamp: 'jlb 2/27/2002 21:07'!
initialize
	super initialize.
	self on: #mouseEnter send: #mouseEnter:for: to: self.! !


!TwoModeButtonMorph class methodsFor: 'example' stamp: 'jlb 7/22/2001 19:57'!
example
	| form |
	" TwoModeButtonMorph example "
	" The file 'close.bmp' should be in the current directory "
	form _ Form fromFileNamed: 'close.bmp' .
	" For this demo, the corners of the button appear to be rounded. Transparencies are coded as black in the bitmap of the button "
	form replaceColor: (Color r: 0.0 g: 0.0 b: 0.004) withColor: Color transparent.
	" open an X button ; click on button to open a Browser "
	^ self new fromImages: form ; 
		target: Browser ;
		actionSelector: #openBrowser ;
		position: 10 @ 10 ;
		openInWorld.! !


!TwoModeButtonMorph class reorganize!
('example' example)
!

